@@ -34,6 +34,7 @@
 #include <syscall.h>
 #endif
 #include <dlfcn.h>
+#include <limits.h>
 
 #include <pthread.h>
 
@@ -87,6 +88,14 @@
 #define PRINTF_ATTRIBUTE(a,b)
 #endif /* HAVE_FUNCTION_ATTRIBUTE_FORMAT */
 
+#ifndef FALL_THROUGH
+# ifdef HAVE_FALLTHROUGH_ATTRIBUTE
+#  define FALL_THROUGH __attribute__ ((fallthrough))
+# else /* HAVE_FALLTHROUGH_ATTRIBUTE */
+#  define FALL_THROUGH
+# endif /* HAVE_FALLTHROUGH_ATTRIBUTE */
+#endif /* FALL_THROUGH */
+
 #define UWRAP_DLIST_ADD(list,item) do { \
 	if (!(list)) { \
 		(item)->prev	= NULL; \
@@ -231,6 +240,9 @@ typedef int (*__libc_getresgid)(gid_t *rgid, gid_t *egid, gid_t *sgid);
 typedef gid_t (*__libc_getegid)(void);
 
 typedef int (*__libc_getgroups)(int size, gid_t list[]);
+#ifdef HAVE___GETGROUPS_CHK
+typedef int (*__libc___getgroups_chk)(int size, gid_t list[], size_t listlen);
+#endif
 
 typedef int (*__libc_setgroups)(size_t size, const gid_t *list);
 
@@ -276,6 +288,9 @@ struct uwrap_libc_symbols {
 #endif
 	UWRAP_SYMBOL_ENTRY(getegid);
 	UWRAP_SYMBOL_ENTRY(getgroups);
+#ifdef HAVE___GETGROUPS_CHK
+	UWRAP_SYMBOL_ENTRY(__getgroups_chk);
+#endif
 	UWRAP_SYMBOL_ENTRY(setgroups);
 #ifdef HAVE_SYSCALL
 	UWRAP_SYMBOL_ENTRY(syscall);
@@ -365,7 +380,14 @@ static pthread_mutex_t libpthread_symbol_binding_mutex = PTHREAD_MUTEX_INITIALIZ
  *********************************************************/
 
 bool uid_wrapper_enabled(void);
+#if ! defined(HAVE_CONSTRUCTOR_ATTRIBUTE) && defined(HAVE_PRAGMA_INIT)
+/* xlC and other oldschool compilers support (only) this */
+#pragma init (uwrap_constructor)
+#endif
 void uwrap_constructor(void) CONSTRUCTOR_ATTRIBUTE;
+#if ! defined(HAVE_DESTRUCTOR_ATTRIBUTE) && defined(HAVE_PRAGMA_FINI)
+#pragma fini (uwrap_destructor)
+#endif
 void uwrap_destructor(void) DESTRUCTOR_ATTRIBUTE;
 
 /*********************************************************
@@ -386,14 +408,30 @@ static void *uwrap_load_lib_handle(enum uwrap_lib lib)
 	int i;
 
 #ifdef RTLD_DEEPBIND
-	flags |= RTLD_DEEPBIND;
+	const char *env_preload = getenv("LD_PRELOAD");
+	const char *env_deepbind = getenv("UID_WRAPPER_DISABLE_DEEPBIND");
+	bool enable_deepbind = true;
+
+	/* Don't do a deepbind if we run with libasan */
+	if (env_preload != NULL && strlen(env_preload) < 1024) {
+		const char *p = strstr(env_preload, "libasan.so");
+		if (p != NULL) {
+			enable_deepbind = false;
+		}
+	}
+
+	if (env_deepbind != NULL && strlen(env_deepbind) >= 1) {
+		enable_deepbind = false;
+	}
+
+	if (enable_deepbind) {
+		flags |= RTLD_DEEPBIND;
+	}
 #endif
 
 	switch (lib) {
 	case UWRAP_LIBNSL:
-		/* FALL TROUGH */
 	case UWRAP_LIBSOCKET:
-		/* FALL TROUGH */
 	case UWRAP_LIBC:
 		handle = uwrap.libc.handle;
 		if (handle == NULL) {
@@ -605,6 +643,17 @@ static int libc_getgroups(int size, gid_t list[])
 	return uwrap.libc.symbols._libc_getgroups.f(size, list);
 }
 
+#ifdef HAVE___GETGROUPS_CHK
+static int libc___getgroups_chk(int size, gid_t list[], size_t listlen)
+{
+	uwrap_bind_symbol_libc(__getgroups_chk);
+
+	return uwrap.libc.symbols._libc___getgroups_chk.f(size,
+							  list,
+							  listlen);
+}
+#endif /* HAVE___GETGROUPS_CHK */
+
 static int libc_setgroups(size_t size, const gid_t *list)
 {
 	uwrap_bind_symbol_libc(setgroups);
@@ -749,7 +798,7 @@ static int uwrap_pthread_create(pthread_t *thread,
 
 	UWRAP_LOCK(uwrap_id);
 
-	args->id->groups = malloc(sizeof(gid_t) * src_id->ngroups);
+	args->id->groups = calloc(src_id->ngroups, sizeof(gid_t));
 	if (args->id->groups == NULL) {
 		UWRAP_UNLOCK(uwrap_id);
 		SAFE_FREE(args->id);
@@ -972,6 +1021,31 @@ static void uwrap_thread_child(void)
 	UWRAP_UNLOCK_ALL;
 }
 
+static unsigned long uwrap_get_xid_from_env(const char *envname)
+{
+	unsigned long xid;
+	const char *env = NULL;
+	char *endp = NULL;
+
+	env = getenv(envname);
+	if (env == NULL) {
+		return ULONG_MAX;
+	}
+
+	if (env[0] == '\0') {
+		unsetenv(envname);
+		return ULONG_MAX;
+	}
+
+	xid = strtoul(env, &endp, 10);
+	unsetenv(envname);
+	if (env == endp) {
+		return ULONG_MAX;
+	}
+
+	return xid;
+}
+
 /*
  * This initializes uid_wrapper with the IDs exported to the environment. Those
  * are normally set after we forked and executed.
@@ -980,62 +1054,61 @@ static void uwrap_init_env(struct uwrap_thread *id)
 {
 	const char *env;
 	int ngroups = 0;
+	unsigned long xid;
 
-	env = getenv("UID_WRAPPER_INITIAL_RUID");
-	if (env != NULL && env[0] != '\0') {
-		UWRAP_LOG(UWRAP_LOG_DEBUG, "Initialize ruid with %s", env);
-		id->ruid = strtoul(env, (char **)NULL, 10);
-		unsetenv("UID_WRAPPER_INITIAL_RUID");
+	/* UIDs */
+	xid = uwrap_get_xid_from_env("UID_WRAPPER_INITIAL_RUID");
+	if (xid != ULONG_MAX) {
+		id->ruid = (uid_t)xid;
 	}
 
-	env = getenv("UID_WRAPPER_INITIAL_EUID");
-	if (env != NULL && env[0] != '\0') {
-		UWRAP_LOG(UWRAP_LOG_DEBUG, "Initalize euid with %s", env);
-		id->euid = strtoul(env, (char **)NULL, 10);
-		unsetenv("UID_WRAPPER_INITIAL_EUID");
+	xid = uwrap_get_xid_from_env("UID_WRAPPER_INITIAL_EUID");
+	if (xid != ULONG_MAX) {
+		id->euid = (uid_t)xid;
 	}
 
-	env = getenv("UID_WRAPPER_INITIAL_SUID");
-	if (env != NULL && env[0] != '\0') {
-		UWRAP_LOG(UWRAP_LOG_DEBUG, "Initalize suid with %s", env);
-		id->suid = strtoul(env, (char **)NULL, 10);
-		unsetenv("UID_WRAPPER_INITIAL_SUID");
+	xid = uwrap_get_xid_from_env("UID_WRAPPER_INITIAL_SUID");
+	if (xid != ULONG_MAX) {
+		id->suid = (uid_t)xid;
 	}
 
-	env = getenv("UID_WRAPPER_INITIAL_RGID");
-	if (env != NULL && env[0] != '\0') {
-		UWRAP_LOG(UWRAP_LOG_DEBUG, "Initialize ruid with %s", env);
-		id->rgid = strtoul(env, (char **)NULL, 10);
-		unsetenv("UID_WRAPPER_INITIAL_RGID");
+	/* GIDs */
+	xid = uwrap_get_xid_from_env("UID_WRAPPER_INITIAL_RGID");
+	if (xid != ULONG_MAX) {
+		id->rgid = (gid_t)xid;
 	}
 
-	env = getenv("UID_WRAPPER_INITIAL_EGID");
-	if (env != NULL && env[0] != '\0') {
-		UWRAP_LOG(UWRAP_LOG_DEBUG, "Initalize egid with %s", env);
-		id->egid = strtoul(env, (char **)NULL, 10);
-		unsetenv("UID_WRAPPER_INITIAL_EGID");
+	xid = uwrap_get_xid_from_env("UID_WRAPPER_INITIAL_EGID");
+	if (xid != ULONG_MAX) {
+		id->egid = (gid_t)xid;
 	}
 
-	env = getenv("UID_WRAPPER_INITIAL_SGID");
-	if (env != NULL && env[0] != '\0') {
-		UWRAP_LOG(UWRAP_LOG_DEBUG, "Initalize sgid with %s", env);
-		id->sgid = strtoul(env, (char **)NULL, 10);
-		unsetenv("UID_WRAPPER_INITIAL_SGID");
+	xid = uwrap_get_xid_from_env("UID_WRAPPER_INITIAL_SGID");
+	if (xid != ULONG_MAX) {
+		id->sgid = (gid_t)xid;
 	}
 
 	env = getenv("UID_WRAPPER_INITIAL_GROUPS_COUNT");
 	if (env != NULL && env[0] != '\0') {
-		ngroups = strtol(env, (char **)NULL, 10);
+		char *endp = NULL;
+		long n;
+
+		n = strtol(env, &endp, 10);
+		if (env == endp) {
+			ngroups = 0;
+		} else if (n > 0 && n < GROUP_MAX_COUNT) {
+			ngroups = (int)n;
+		}
 		unsetenv("UID_WRAPPER_INITIAL_GROUPS_COUNT");
 	}
 
-	if (ngroups > 0 && ngroups < GROUP_MAX_COUNT) {
+	if (ngroups > 0) {
 		int i = 0;
 
 		id->ngroups = 0;
 
 		free(id->groups);
-		id->groups = malloc(sizeof(gid_t) * ngroups);
+		id->groups = calloc(ngroups, sizeof(gid_t));
 		if (id->groups == NULL) {
 			UWRAP_LOG(UWRAP_LOG_ERROR,
 				  "Unable to allocate memory");
@@ -2081,6 +2154,30 @@ int getgroups(int size, gid_t *list)
 	return uwrap_getgroups(size, list);
 }
 
+#ifdef HAVE___GETGROUPS_CHK
+static int uwrap___getgroups_chk(int size, gid_t *list, size_t listlen)
+{
+	if (size * sizeof(gid_t) > listlen) {
+		UWRAP_LOG(UWRAP_LOG_DEBUG, "Buffer overflow detected");
+		abort();
+	}
+
+	return uwrap_getgroups(size, list);
+}
+
+int __getgroups_chk(int size, gid_t *list, size_t listlen);
+
+int __getgroups_chk(int size, gid_t *list, size_t listlen)
+{
+	if (!uid_wrapper_enabled()) {
+		return libc___getgroups_chk(size, list, listlen);
+	}
+
+	uwrap_init();
+	return uwrap___getgroups_chk(size, list, listlen);
+}
+#endif /* HAVE___GETGROUPS_CHK */
+
 #if (defined(HAVE_SYS_SYSCALL_H) || defined(HAVE_SYSCALL_H)) \
     && (defined(SYS_setreuid) || defined(SYS_setreuid32))
 static long int uwrap_syscall (long int sysno, va_list vp)
